#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Date: 2006-09-24 14:01:41 -0400 (Sun, 24 Sep 2006) $'.split()[1]
__author__ = 'Kurt Schwehr'

__doc__='''

Tools to generate python code to serialize/deserialize messages
between python and ais binary.  Trying to be as inline as possible, so
no XML on the fly like in ais-py.  

serialize: python to ais binary
deserialize: ais binary to python

The generated code uses translators.py, binary.py, and aisstring.py
which should be packaged with the resulting files.

@requires: U{lxml<http://codespeak.net/lxml/>}
@requires: U{epydoc<http://epydoc.sourceforge.net/>} >= 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>} >= 1.2

@author: U{'''+__author__+'''<http://schwehr.org/>}
@version: ''' + __version__ +'''
@copyright: 2006
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser
@since: 2006-Sep-24
@status: under development
@organization: U{CCOM<http://ccom.unh.edu/>}
@license: GPL v2

@todo: add a link to generated doc string to bring up the html for the pretty version
@todo: write a separate validation script that distinguishes standard messages and bin messages
@todo: make sure binary is only used in AIS ITU messages and not within the binary messages!
@todo: arrays
@bug: NOT complete
'''

import sys, os
from decimal import Decimal
from lxml import etree 

def suggestType(name,curType,printout=True):
    '''
    Try to suggest a type name if one did not work.

    @param printout: if true, write a suggestion to stdout.

    >>> suggestType('myFieldName','unsigned int')
    Recommend switching "unsigned int" to "uint" for field "myFieldName"
    'uint'

    >>> suggestType('JohnWarfon','yoyodyne')
    Sorry!  No recommendation available for bad type "yoyodyne" for field "JohnWarfon"
    '''
    newType = None
    if curType.lower()=='unsigned int':
	newType = 'uint'
    elif curType.lower()=='unsigned decimal':
	newType = 'udecimal'

    if printout:
	if None != newType:
	    print 'Recommend switching "'+curType+'" to "'+newType+'" for field "'+name+'"'
	else:
	    print 'Sorry!  No recommendation available for bad type "'+curType+'" for field "'+name+'"'
    return newType


def hasSubtag(et,subtag):
    '''
    @return: true if the tag a sub tag with name subtag
    '''
    if 0<len(et.xpath(subtag)): return True
    return False


#def writeBeginning(o,aisBinMsgET):
def writeBeginning(o):
    '''
    Write the doc string header for the message file

    param o: Open output file to write code to.
    param msgET: element tree for the ais message definition.  
    Must be pre-expanded with the expandais.py command.
    '''
    import datetime
    d = datetime.datetime.utcnow()
    dateStr = str(d.year)+'-'+("%02d" %d.month)+'-'+("%02d"%d.day)

    # FIX: what to do for __version__, @since, etc?
    # Need to pass in info about the source file, etc.

    # Minor trickery to get svn to ignore the keywords in the next few lines
    o.write('''#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Da'''+'''te: '''+dateStr+''' $'.split()[1]
__author__ = 'xmlbinmsg'

__doc__=\'\'\'

Autogenerated python functions to serialize/deserialize binary messages.

Generated by: '''+__file__+'''

Need to then wrap these functions with the outer AIS packet and then
convert the whole binary blob to a NMEA string.  Those functions are
not currently provided in this file.

serialize: python to ais binary
deserialize: ais binary to python

The generated code uses translators.py, binary.py, and aisstring.py
which should be packaged with the resulting files.

''')

    o.write('''
@requires: U{epydoc<http://epydoc.sourceforge.net/>} > 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>}

@author: \'\'\'+__author__+\'\'\'
@version: \'\'\' + __version__ +\'\'\'
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser
@status: under development
@license: Generated code has no license
\'\'\'

import sys
from decimal import Decimal
from BitVector import BitVector

import binary, aisstring

# FIX: check to see if these will be needed
TrueBV  = BitVector(bitstring="1")
"Why always rebuild the True bit?  This should speed things up a bunch"
FalseBV = BitVector(bitstring="0")
"Why always rebuild the False bit?  This should speed things up a bunch"


''')
    return

def generatePython(infile,outfile, prefixName=False):
    '''
    @param infile: xml ais binary message definition file
    @param outfile: where to dump the python code
    '''

    aisMsgsET = etree.parse(infile).getroot()

    o = file(outfile,'w')
    os.chmod(outfile,0755)

    writeBeginning(o)

    for msgET in aisMsgsET:
	if msgET.tag != 'message': continue
	print msgET.tag, msgET.attrib['name']

	if len(msgET.xpath('include-struct')) > 0:
	    sys.exit("ERROR: cannot handle xml that still has include-struct tags.\n  Please use expandais.py.")
	buildEncode(o,msgET,prefixName=prefixName)
	buildDecode(o,msgET,prefixName=prefixName)
	buildDecodeParts(o,msgET,prefixName=prefixName) # functions that only decode one field
	buildPrint(o,msgET,prefixName=prefixName)

    o.write('\n\n######################################################################\n')
    o.write('# UNIT TESTING\n')
    o.write('######################################################################\n')
    o.write('import unittest\n')

    for msgET in aisMsgsET:
	if msgET.tag != 'message': continue
	print 'Building unit tests for message ...', msgET.attrib['name']

	buildUnitTest(o,msgET, prefixName=prefixName)
	
    for msgET in aisMsgsET:
	if msgET.tag != 'message': continue
	buildMain(o,msgET, prefixName=prefixName)

    return 

######################################################################
# SIMPLE PRINT
######################################################################

def getMaxFieldNameLen(msgET):
    '''Get the maximum string length of any field name'''
    maxStrLen=0
    for field in msgET.xpath('field'):
	fieldLen = len(field.attrib['name'])
	if fieldLen>maxStrLen: maxStrLen = fieldLen
    return maxStrLen

def padStrRight(aStr,strlen):
    '''Pad a string out to the length requested with spaces out to the right'''
    return aStr + ' '*(strlen-len(aStr))

def buildPrint(o,msgET, verbose=False, prefixName=False):
    '''
    Write a simple in order print for the resulting dictionary.

    param o: open file where resulting code will be written
    param msgET: Element Tree starting at a message node
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating print for',name # FIX: verbose?
    funcName = 'printFields'
    if prefixName: funcName = name+'PrintFields'
    
    o.write('def '+funcName+'(params, out=sys.stdout, format=\'std\'):\n')

    ########################################
    # doc string
    o.write("\t'''Print a "+name+" message to stdout.\n\n")
    o.write('\tFields in params:\n')
    for field in msgET.xpath('field'):
	desc = field[0].text.replace('\n',' ') # get ride of new lines
	o.write('\t  - '+field.attrib['name']+'('+field.attrib['type']+'): '+desc) # give the description
	if len(field.xpath("required")) == 1:
	    o.write(' (field automatically set to "'+field.xpath("required")[0].text+'")')

	o.write('\n')
    o.write('\t@param params: Dictionary of field names/values.  \n')
    o.write('\t@param out: File like object to write to\n')
	
    o.write("\t@rtype: stdout\n")
    o.write("\t@return: text to out\n")
    o.write("\t'''\n\n")

    ########################################
    # Actually build the code

    o.write('\tif \'std\'==format:\n')
    o.write('\t\tout.write("'+name+':\\n")\n')

    if verbose: print 'number of fields = ', len(msgET.xpath('field'))

    # +1 for the ':'
    maxFieldLen = 1 + getMaxFieldNameLen(msgET)
    

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	    #print 'required set for',name,'to',required
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'
	else:
	    if verbose: print 'Processing field ...',name,'(',type+' ',numbits,')'

	if 1==arraylen or type=='aisstr6':
	    o.write('\t\tout.write("\t'+padStrRight(name+':',maxFieldLen)+'  "+str(params[\''+name+'\'])+"\\n")\n')
	    # FIX: elif aisstr6 strip @@@ and then print
	else:
	    sys.exit ('FIX: handle arrays in the buildPrint func')



    o.write('\n\treturn # Nothing to return\n\n')

######################################################################
# ENCODERS
######################################################################
def encodeBool(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for boolean variables
    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: bool, etc.
    @type numbits: int = 1
    @param numbits: How many bits per unit datum (must be 1 for bools)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many bools will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    
    if verbose: print 'bool encode',name,': unvail=',unavailable

    assert type.lower()=='bool'
    assert numbits==1
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type=bool)\n')
    if None != required:
	assert type(required)==bool
	if required: o.write('\t\tbvList.append(TrueBV)\n')
	else: o.write('\t\tbvList.append(FalseBV)\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tif params["'+name+'"]: bvList.append(TrueBV)\n')
	o.write('\telse: bvList.append(FalseBV)\n')
    else: # Have a default value that can be filled in
	assert type(unavailable)==bool
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tif params["'+name+'"]: bvList.append(TrueBV)\n')
	o.write('\t\telse: bvList.append(FalseBV)\n')
	o.write('\telse:\n')
	if unavailable: o.write('\t\tbvList.append(TrueBV)\n')
	else: o.write('\t\tbvList.append(FalseBV)\n')
    if verbose: o.write('\n')

def encodeUInt(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert type=='uint'
    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal='+str(required)+'),'+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal=params[\''+name+'\']),'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal=params[\''+name+'\']'+'),'+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal='+str(unavailable)+'),'+str(numbits)+'))\n')

    if verbose: o.write('\n')


def encodeFloat(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for IEEE float variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits==32 # Force by the IEEE spec
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.float2bitvec('+str(required)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(binary.float2bitvec(params[\''+name+'\']))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.float2bitvec(params[\''+name+'\']'+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.float2bitvec('+str(unavailable)+'))\n')

    if verbose: o.write('\n')


def encodeAisstr6(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for aisstr6 variables.  Generally are arrays.
    @bug: do we need to optionally check for a valid string?

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits==6 # Each character must be 6 bits
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(aisstring.encode(\''+str(required)+'\','+str(numbits*arraylen)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(aisstring.encode(params[\''+name+'\'],'+str(numbits*arraylen)+'))\n')
    else: # Have a default value that can be filled in
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(aisstring.encode(params[\''+name+'\'],'+str(numbits*arraylen)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(aisstring.encode(\''+str(unavailable)+'\','+str(numbits*arraylen)+'))\n')

    if verbose: o.write('\n')


def encodeInt(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for signed integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many signed ints will there be?  FIX: handle variable
    @type unavailable: number or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.bvFromSignedInt('+str(required)+','+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return


    if None==unavailable:
	o.write('\tbvList.append(binary.bvFromSignedInt(params[\''+name+'\'],'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.bvFromSignedInt(params[\''+name+'\']'+','+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.bvFromSignedInt('+str(unavailable)+','+str(numbits)+'))\n')

    if verbose: o.write('\n')



# FIX: Ummm... why am I passing the type?  I guess it makes the one print statement easier
def encodeDecimal(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False, scale=None):
    '''
    Build the encoder for signed decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: decimal
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many decimals will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeDecimal:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    # FIX: optimize to not emit the scaling when it is not needed... or tell the user to switch to an int!
    if None == scale: 
	print 'WARNING: if you are not scaling, then you probably want to use an int instead!'
	print 'Beware canadians bearing travel videos'
	scale='1'

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.bvFromSignedInt('+str(int(Decimal(required)*Decimal(scale)))+','+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return

    # FIX: can I get rid of the Decimal around params?
    if None==unavailable:
	o.write('\tbvList.append(binary.bvFromSignedInt(int(Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')),'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.bvFromSignedInt(int(Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')),'+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.bvFromSignedInt('+str(int(Decimal(unavailable)*Decimal(scale)))+','+str(numbits)+'))\n')

    if verbose: o.write('\n')



def encodeUDecimal(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False, scale=None):
    '''
    Build the encoder for signed decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: decimal
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many decimals will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeDecimal:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable
    assert type=='udecimal'
    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    # FIX: optimize to not emit the scaling when it is not needed... or tell the user to switch to an int!
    if None == scale: 
	print 'WARNING: if you are not scaling, then you probably want to use an int instead!'
	print 'Beware canadians bearing travel videos'
	scale='1'

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	assert(0<=required)
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal='+str(int(Decimal(required)*Decimal(scale)))+'),'+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return

    # FIX: can I get rid of the Decimal around params?
    if None==unavailable:
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal=int((Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')))),'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal=int((Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')))),'+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal=int('+str(int(Decimal(unavailable)*Decimal(scale)))+')),'+str(numbits)+'))\n')

    if verbose: o.write('\n')



def encodeBinary(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False, scale=None):
    '''
    Build the encoder for binary variables.  This is just a pass through.
    This is used for the ais binary message wrapper (e.g. msg 8).  Do
    not use it within binary messages.

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: binary
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..1024 or so)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many decimals will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encode'+name+':',type,numbits,'Req:',required,'alen:',arraylen,unavailable
    assert type=='binary'
    assert numbits>=1 and numbits<=1024
    assert (None == required)  # don't allow this
    assert (None == unavailable) # don't allow this

    if arraylen != 1: assert False # Do not handle arrays.  Arrays of bits is just not necessary.
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    # FIX: can I get rid of the Decimal around params?
    o.write('\tbvList.append(params[\''+name+'\'])\n')  # Just pass it through

    if verbose: o.write('\n')


######################################################################
# DECODERS
######################################################################

def decodeBool(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False, decodeOnly=False):
    '''
    Build the decoder for boolean variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type startindex: int
    @param startindex: bit that begins the bool(s)
    @type numbits: int = 1
    @param numbits: How many bits per unit datum (must be 1 for bools)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many bools will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert(type=='bool')
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    #int(startindex); int(numbits)  # Make sure it is a number
    assert numbits==1
    assert arraylen == 1 # FIX... handle arrays

    if None != required:
	assert type(required)==bool
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	if required: o.write('True\n')
	else: o.write('False\n')
	if not decodeOnly: o.write('\n')
	return int(startindex)+int(numbits)

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('bool(int('+bv+'['+str(startindex)+':'+str(startindex+int(numbits)*int(arraylen))+']))')
    if not decodeOnly: o.write('\n')

    return int(startindex)+int(numbits)


def decodeUInt(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False, decodeOnly=False):
    '''
    Build the decoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, etc.
    @type startindex: int
    @param startindex: bit that begins the uint(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: int or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: int or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1
    if not decodeOnly: verbose=False

    if None != required:
	int(required) # Make sure required is a number
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(str(required))
	if not decodeOnly: o.write('\n')
	return startindex+numbits

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('int('+bv+'['+str(startindex)+':'+str(startindex+int(numbits)*int(arraylen))+'])')
    if not decodeOnly: o.write('\n')
    if verbose: o.write('\n')
    
    return startindex+numbits


def decodeInt(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False, decodeOnly=False):
    '''
    Build the decoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: int
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: int or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: int or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='int'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1

    if None != required:
	int(required) # Make sure required is a number
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(str(required))
	if not decodeOnly: o.write('\n')
	return end

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('binary.signedIntFromBV('+bv+'['+str(startindex)+':'+str(end)+'])')
    if not decodeOnly: o.write('\n')
    if verbose: o.write('\n')
    
    return end

def decodeFloat(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False, decodeOnly=False):
    '''
    Build the decoder for IEEE float variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: int
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: float or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: float or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='float'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1

    if None != required:
	float(required) # Make sure required is a number
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(str(required))
	if not decodeOnly: o.write('\n')
	if verbose: o.write('\n')
	return end

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('binary.bitvec2float('+bv+'['+str(startindex)+':'+str(end)+'])')
    if not decodeOnly: o.write('\n')
    
    return end


def decodeAisstr6(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
		  bv='bv',dataDict='r',verbose=False, decodeOnly=False):
    '''
    Build the decoder for aisstr6 variables.  Generally arrays.
    @bug: FIX: validate strings??
    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'aisstr6'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: restricted str or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: restricted str or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='aisstr6'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen >= 1 # FIX... handle arrays
    assert numbits>=1

    if None != required:
	float(required) # Make sure required is a number
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(required)
	if not decodeOnly: o.write('\n')
	return end

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('aisstring.decode('+bv+'['+str(startindex)+':'+str(end)+'])')
    if not decodeOnly: o.write('\n')
    
    return end


def decodeDecimal(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False,scale=None, decodeOnly=False):
    '''
    Build the decoder for signed decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'decimal'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: Decimal or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='decimal'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1 and numbits <= 32

    if None == scale: scale='1' # Warning about this was in the encode section

    if None != required:
	Decimal(required) # Make sure required is a number
	if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(str(Decimal(required))+'/Decimal(\''+scale+'\')')
	if not decodeOnly: o.write('\n')
	return end

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('Decimal(binary.signedIntFromBV('+bv+'['+str(startindex)+':'+str(end)+']))/Decimal(\''+scale+'\')')
    if not decodeOnly: o.write('\n')
    
    return end




def decodeUDecimal(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False,scale=None, decodeOnly=False):
    '''
    Build the decoder for unsigned decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'udecimal'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: Decimal or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='udecimal'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1 and numbits <= 32

    if None == scale: scale='1' # Warning about this was in the encode section

    if None != required:
	assert (Decimal(required)>=0.) # Make sure required is a number and not negative
	if not decodeOnly:	o.write('\t'+dataDict+'[\''+name+'\']=')
	o.write(str(Decimal(required))+'/Decimal(\''+scale+'\')')
	if not decodeOnly: o.write('\n')
	return end

    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write('Decimal(int('+bv+'['+str(startindex)+':'+str(end)+']))/Decimal(\''+scale+'\')')
    if not decodeOnly: o.write('\n')
    
    return end


def decodeBinary(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False,scale=None, decodeOnly=False):
    '''
    Build the decoder for unsigned decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'udecimal'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: Decimal or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @type decodeOnly: bool
    @param decodeOnly: Set to true to only get the code for decoding
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='binary'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1 and numbits <= 1024 # What is good max?
    # FIX: assert not required and not an array an not unavailable


    if not decodeOnly: o.write('\t'+dataDict+'[\''+name+'\']=')
    o.write(bv+'['+str(startindex)+':'+str(end)+']')
    if not decodeOnly: o.write('\n')
    
    return end



######################################################################
# THE REST
######################################################################



def buildTestParamFunc(o,msgET, verbose=False, prefixName=False):
    '''Scrape the testvalues to make a basic param

    @bug: FIX: make this create a dictionary that sits in the overall namespace and spit out deep copies?
    '''
    name = msgET.attrib['name']

    if prefixName: o.write('def '+name+'TestParams():\n')
    else: o.write('def testParams():\n')
    o.write("\t'''Return a params file base on the testvalue tags.\n\t@rtype: dict\n\t@return: params based on testvalue tags\n\t'''\n")
    o.write('\tparams = {}\n')
    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
	if verbose: print 'buildTestParamFunc ...',name,type
	val = None
	if hasSubtag(field,'testvalue') and hasSubtag(field,'required'):
	    print 'ERROR: can not have both test value and required tags in the same field'
	    assert(False)
	if hasSubtag(field,'testvalue'):
	    val = field.xpath('testvalue')[0].text
	else:
	    if not hasSubtag(field,'required'):
		sys.exit("ERROR: missing required or testvalue for field: "+name)
	    val = field.xpath('required')[0].text
	if verbose: print 'buildTestParamFunc for field '+name+' ...',type,val
	o.write('\tparams[\''+name+'\'] = ')
	if type=='bool': 
	    if val=='1' or val.lower=='true': val = 'True'
	    else: val = 'False'
	    o.write(val)
	elif type in ('uint','int','float'):
	    o.write(val)
	elif type in ('decimal','udecimal'):
	    o.write('Decimal(\''+val+'\')')

	elif type in ('aisstr6'):
	    o.write('\''+val+'\'')
	elif type in ('binary'):
	    o.write('BitVector(bitstring=\''+val+'\')')
	else:
	    print 'ERROR: type not handled ...',type,'  (found in the ',name,' field).  Time to buy more coffee'
	    suggestType(name,type)
	    assert(False)

	o.write('\n')


    o.write('\n\treturn params\n\n')

def buildUnitTest(o,msgET, verbose=False, prefixName=False):
    '''
    Write the unittests for a message

    param o: open file where resulting code will be written
    param msgET: Element Tree starting at a message node
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    buildTestParamFunc(o,msgET, prefixName=prefixName)

    o.write('class Test'+name+'(unittest.TestCase):\n')
    o.write("\t'''Use testvalue tag text from each type to build test case the "+name+" message'''\n")
    o.write('\tdef testEncodeDecode(self):\n\n')
    if prefixName:
	o.write('\t\tparams = '+name+'TestParams()\n')
	o.write('\t\tbits   = '+name+'Encode(params)\n')
	o.write('\t\tr      = '+name+'Decode(bits)\n\n')
    else:
	o.write('\t\tparams = testParams()\n')
	o.write('\t\tbits   = encode(params)\n')
	o.write('\t\tr      = decode(bits)\n\n')

    o.write('\t\t# Check that each parameter came through ok.\n')
    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
	if type in ('bool','uint','int','aisstr6','binary'):
	    o.write('\t\tself.failUnlessEqual(r[\''+name+'\'],params[\''+name+'\'])\n')
	else:
	    # float, decimal, udecimal
	    # FIX: look up the decimal places if decimal
	    places = '7'
	    if hasSubtag(field,'decimalplaces'): places = field.xpath('decimalplaces')[0].text
	    o.write('\t\tself.failUnlessAlmostEqual(r[\''+name+'\'],params[\''+name+'\'],'+places+')\n')
	


def buildEncode(o,msgET, verbose=False, prefixName=False):
    '''
    Write the encoder/decoder for a message

    http://jaynes.colorado.edu/PythonIdioms.html

    param o: open file where resulting code will be written
    param msgET: Element Tree starting at a message node
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating encoder for',name # FIX: verbose?
    funcName = 'encode'
    if prefixName: funcName = name+'Encode'
    o.write('def '+funcName+'(params, validate=False):\n')

    ########################################
    # doc string
    o.write("\t'''Create a "+name+" binary message payload to pack into an AIS Msg "+name+".\n\n")
    o.write('\tFields in params:\n')
    for field in msgET.xpath('field'):
	desc = field[0].text.replace('\n',' ') # get ride of new lines
	o.write('\t  - '+field.attrib['name']+'('+field.attrib['type']+'): '+desc) # give the description
	if len(field.xpath("required")) == 1:
	    o.write(' (field automatically set to "'+field.xpath("required")[0].text+'")')

	o.write('\n')
    o.write('\t@param params: Dictionary of field names/values.  Throws a ValueError exception if required is missing\n')
    o.write('\t@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.\n')
	
    o.write("\t@rtype: BitVector\n")
    o.write("\t@return: encoded binary message (for binary messages, this needs to be wrapped in a msg 8\n")
    o.write("\t@note: The returned bits may not be 6 bit aligned.  It is up to you to pad out the bits.\n")
    o.write("\t'''\n\n")

    ########################################
    # Actually build the code

    o.write('\tbvList = []\n')

    if verbose: print 'number of fields = ', len(msgET.xpath('field'))

    dynamicArrays = False # Set to true when positioning must be calculated

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	    #print 'required set for',name,'to',required
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'
	else:
	    if verbose: print 'Processing field ...',name,'(',type+' ',numbits,')'

	if   type=='bool'   : encodeBool   (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='uint'   : encodeUInt   (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='int'    : encodeInt    (o,name,type,numbits,required,arraylen,unavailable)	    
	elif type=='float'  : encodeFloat  (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='aisstr6': encodeAisstr6(o,name,type,numbits,required,arraylen,unavailable)	    
        elif type=='decimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    encodeDecimal(o,name,type,numbits,required,arraylen,unavailable,scale=scale)
        elif type=='udecimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    encodeUDecimal(o,name,type,numbits,required,arraylen,unavailable,scale=scale)
	elif type=='binary': encodeBinary(o,name,type,numbits,required,arraylen,unavailable)
        else: 
	    print 'WARNING: In buildEncode - Unhandled field type for',name,'...',type
	    suggestType (name,type)
	    assert False

#    o.write('\n\tbv=binary.joinBV(bvList)\n')
#    o.write('\n\tbvLen=len(bv)\n')
#    o.write('\n\tif bvLen%6!=0:\n')
#    o.write('\n\t    bv = bv + BitVector(size=bvLen%6)  # \n')
#    o.write('\n\treturn bv\n\n')
    o.write('\n\treturn binary.joinBV(bvList)\n\n')


######################################################################
# DECODER ONE PART AT A TIME - if only using a small part, save some cycles!


#, msgDict=None):
def buildDecodeParts(o,msgET, verbose=False, prefixName=False):

    '''
    Write the decoder for a message

    param o: open file where resulting code will be written
    type msgET: elementtree
    param prefixName: if True, put the name of the message on the functions.
    param msgET: Element Tree starting at a message node
    return: None

    @todo: FIX: check for a dac,fid, or efid.  If exists, then this is an AIS Msg 8 payload
    @todo: May want to take a dictionary of already decoded fields to speed things that need prior info
    for things like variable length arrays
    '''

    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating partial decode functions for',name
    
    baseName = name+'Decode'
    if not prefixName: baseName = 'decode'

    startindex = 0 # Where we are in the bitvector... FIX: what about variable length jobs?

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']

	o.write('def '+baseName+name+'(bv, validate=False):\n')
	# Follow the same convention of decoding into a dict so that code is the same
	#o.write('\tr={};')
	o.write('\treturn ')

        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'

	assert None!=startindex
	if verbose: print 'startindex',startindex
	if   type=='bool'   : startindex = decodeBool   (o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)
	elif type=='uint'   : startindex = decodeUInt   (o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)
	elif type=='int'    : startindex = decodeInt    (o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)
	elif type=='float'  : startindex = decodeFloat  (o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)
	elif type=='aisstr6': startindex = decodeAisstr6(o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)
	elif type=='binary' : startindex = decodeBinary (o,name,type,startindex,numbits,required,arraylen,unavailable,decodeOnly=True)	    

        elif type=='decimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    startindex = decodeDecimal(o,name,type,startindex, numbits,required,arraylen,unavailable,scale=scale,decodeOnly=True)
        elif type=='udecimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    startindex = decodeUDecimal(o,name,type,startindex, numbits,required,arraylen,unavailable,scale=scale,decodeOnly=True)

        else: 
	    print 'WARNING: In buildDecode - Unhandled field type for',name,'...',type
	    suggestType (name,type)
	    assert False

	#o.write('\treturn(r[\''+name+'\'])\n\n')
	o.write('\n\n')


######################################################################
# DECODER RING

def buildDecode(o,msgET, verbose=False, prefixName=False):
    '''
    Write the decoder for a message

    param o: open file where resulting code will be written
    type msgET: elementtree
    param msgET: Element Tree starting at a message node
    return: None

    @todo: FIX: check for a dac,fid, or efid.  If exists, then this is an AIS Msg 8 payload
    '''

    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating decoder for',name
    funcName = 'decode'
    if prefixName: funcName = name+'Decode'
    o.write('def '+funcName+'(bv, validate=False):\n')

    ########################################
    # doc string
    o.write("\t'''Unpack a "+name+" message \n\n")
    o.write('\tFields in params:\n')
    for field in msgET.xpath('field'):
	desc = field[0].text.replace('\n',' ') # get ride of new lines
	o.write('\t  - '+field.attrib['name']+'('+field.attrib['type']+'): '+desc) # give the description
	if len(field.xpath("required")) == 1:
	    o.write(' (field automatically set to "'+field.xpath("required")[0].text+'")')

	o.write('\n')
    o.write('\t@type bv: BitVector\n')
    o.write('\t@param bv: Bits defining a message\n')
    o.write('\t@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.\n')
	
    o.write("\t@rtype: dict\n")
    o.write("\t@return: params\n")
    o.write("\t'''\n\n")


    o.write('\t#Would be nice to check the bit count here..\n')
    o.write('\t#if validate:\n')
    o.write('\t#\tassert (len(bv)==FIX: SOME NUMBER)\n')


    ########################################
    # Actually build the code

    o.write('\tr = {}\n')


    if verbose: print 'number of fields = ', len(msgET.xpath('field'))

    dynamicArrays = False # Set to true when positioning must be calculated

    startindex = 0 # Where we are in the bitvector... FIX: what about variable length jobs?

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	    #print 'required set for',name,'to',required
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'
	else:
	    if verbose: print 'Processing field ...',name,'(',type+' ',numbits,')'

	assert None!=startindex
	if verbose: print 'startindex',startindex
	if   type=='bool'   : startindex = decodeBool   (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='uint'   : startindex = decodeUInt   (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='int'    : startindex = decodeInt    (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='float'  : startindex = decodeFloat  (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='aisstr6': startindex = decodeAisstr6(o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='binary' : startindex = decodeBinary (o,name,type,startindex,numbits,required,arraylen,unavailable)	    

        elif type=='decimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    #print 'pre call...'
	    #print '  required: "'+str(required)+'"  scale: "'+str(scale)+'"  unavail:', unavailable
	    startindex = decodeDecimal(o,name,type,startindex, numbits,required,arraylen,unavailable,scale=scale)
        elif type=='udecimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    #print 'pre call...'
	    #print '  required: "'+str(required)+'"  scale: "'+str(scale)+'"  unavail:', unavailable
	    startindex = decodeUDecimal(o,name,type,startindex, numbits,required,arraylen,unavailable,scale=scale)

        else: 
	    print 'WARNING: In buildDecode - Unhandled field type for',name,'...',type
	    suggestType (name,type)
	    assert False


	if None==startindex: print 'FIX: here. drat.  treat me right'
	assert None!=startindex


    o.write('\treturn r\n\n')



#def getPythonType(aType):
#    '''
#    Translate a 
#    @rtype: str
#    @return: name of the python type
#    '''

aisType2pythonType={
'bool':'Bool',
'uint':'int',
'int':'int',
'udecimal':'Decimal',
'decimal':'Decimal',
'aisstr6':'str',
'binary':'str',
'float':'Float',
}

import copy
aisType2optParseType=copy.deepcopy(aisType2pythonType)
aisType2optParseType['bool']='int' # FIX: make these appear as a flag variable
aisType2optParseType['aisstr6']='string'
aisType2optParseType['aisstr6']='string'
aisType2optParseType['binary']='string'
aisType2optParseType['udecimal']='string'  # FIX: Is this a good choice?
aisType2optParseType['decimal']='string'
aisType2optParseType['float']='float'


def buildMain(o, msgET, prefixName=False):
    assert None != msgET
    assert msgET.tag=='message'
    msgName = msgET.attrib['name']

    prefix=''
    if prefixName: prefix=msgName

    o.write('''
############################################################
if __name__=='__main__':

    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)

    #sys.exit('EARLY EXIT - FIX: remove')
    myparser.add_option('--doc-test',dest='doctest',default=False,action='store_true',
                        help='run the documentation tests')
    myparser.add_option('--unit-test',dest='unittest',default=False,action='store_true',
                        help='run the unit tests')
    myparser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
                        help='Make the test output verbose')


    myparser.add_option('-d','--decode',dest='doDecode',default=False,action='store_true',
                        help='decode a "'''+msgName+'''" AIS message')
    myparser.add_option('-e','--encode',dest='doEncode',default=False,action='store_true',
                        help='encode a "'''+msgName+'''" AIS message')

		 
    # FIX: remove nmea from binary messages.  No way to build the whole packet?
    # FIX: or build the surrounding msg 8 for a broadcast?
    typeChoices = ('binary','nmeapayload','nmea') # FIX: what about a USCG type message?
    myparser.add_option('-t','--type',choices=typeChoices,type='choice',dest='ioType'
			,default='nmeapayload'
			,help='What kind of string to expect ('+', '.join(typeChoices)+') [default: %default]')

''')

    #print 'here...', msgName, prefixName
    for field in msgET.xpath('field'):
	name      = field.attrib['name']
	fieldType = field.attrib['type']
	if hasSubtag(field,'required'):
	    print 'skipping required field ...',name,fieldType
	    continue
	#print 'there',name,fieldType
	o.write('    myparser.add_option(\'--')
	if prefixName: o.write(msgName+'-')
	o.write(name+'-field\', dest=\''+name+'Field\'')
	if hasSubtag(field,'unavailable'):
	    val = field.xpath('unavailable')[0].text
	    o.write(',default=')
	    if fieldType in ('uint','int','float'): 
		o.write(val)
	    elif fieldType in ('decimal','udecimal'):
		o.write('Decimal(\''+val+'\')')
	    elif fieldType in ('aisstr6','bitvector'):
		o.write('\''+val+'\'')
	o.write('\n\t,help=\'Field parameter value [default: %default]\'')
	o.write('\n\t,metavar=\''+fieldType+'\'')
	o.write('\n\t,type=\''+aisType2optParseType[fieldType]+'\'')
	o.write('\n\t)\n')


    o.write('''
    (options,args) = myparser.parse_args()
    success=True

    if options.doctest:
	import os; print os.path.basename(sys.argv[0]), 'doctests ...',
	sys.argv= [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	import doctest
	numfail,numtests=doctest.testmod()
	if numfail==0: print 'ok'
	else: 
	    print 'FAILED'
	    success=False

    if not success:
	sys.exit('Something Failed')

    del success # Hide success from epydoc

    if options.unittest:
	sys.argv = [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	unittest.main()
''')


    ##############################
    # encode
    ##############################
    # FIX: make everything tabs.  Grr.
    o.write('\n    if options.doEncode:\n')
    o.write('\t\t# First make sure all non required options are specified\n')
    for field in msgET.xpath('field'):
	name      = field.attrib['name']
	fieldType = field.attrib['type']
	varName = prefix+name+'Field'
	if not hasSubtag(field,'required'):
	    o.write('\t\tif None==options.'+varName+': myparser.error("missing value for '+varName+'")\n')

    # Build dict
    o.write('\t\tmsgDict={\n')
    for field in msgET.xpath('field'):
	name    = field.attrib['name']
	varName = prefix+name+'Field'
	if hasSubtag(field,'required'):
	    o.write('\t\t\t\''+name+'\': \''+field.xpath('required')[0].text+'\',\n')
	else:
	    o.write('\t\t\t\''+name+'\': options.'+varName+',\n')
    o.write('\t\t}\n')

    encodeFunction = 'encode'
    if prefixName: encodeFunction = msgName+'Encode'
    o.write('\t\tbits = '+encodeFunction+'(msgDict)\n')
    o.write('\t\tif \'binary\'==options.ioType: print str(bits)\n')
    o.write('\t\telif \'nmeapayload\'==options.ioType:\n')
    # FIX: figure out if this might be necessary at compile time
    o.write('\t\t\tprint "bitLen",len(bits)\n')
    o.write('\t\t\tbitLen=len(bits)\n')
    o.write('\t\t\tif bitLen%6!=0:\n')
    #o.write('\t\t\t    print "padding...",bitLen%6\n')
    o.write('\t\t\t    bits = bits + BitVector(size=(6 - (bitLen%6)))  # Pad out to multiple of 6\n')
    #o.write('\t\t\tprint "bitLen",len(bits)\n')
    o.write('\t\t\tprint "result:",binary.bitvectoais6(bits)[0]\n')


    # FIX: Do not emit this option for the binary message payloads.  Does not make sense.
    #o.write('\t\tif \'nmea\'==options.ioType: print binarybits\n')
    o.write('\t\telif \'nmea\'==options.ioType: sys.exit("FIX: need to implement this capability")\n')
    o.write('\t\telse: sys.exit(\'ERROR: unknown ioType.  Help!\')\n')


    ##############################
    # decode all
    ##############################
    decodeFunction = 'decode'
    printFields='printFields'
    if prefixName: 
	decodeFunction = msgName+'Decode'
	printFields    = msgName+'PrintFields'
    o.write('\n    if options.doDecode:\n')
    o.write('\t\tfor msg in args:\n')
    o.write('\t\t\tif \'binary\'==options.ioType:\n')
    o.write('\t\t\t    bv = BitVector(bitstring=msg)\n')
    o.write('\t\t\t    '+printFields+'('+decodeFunction+'(bv))\n')

    o.write('\t\t\telif \'nmeapayload\'==options.ioType:\n')
    o.write('\t\t\t    bv = binary.ais6tobitvec(msg)\n')
    o.write('\t\t\t    '+printFields+'('+decodeFunction+'(bv))\n')

    o.write('\t\t\telif \'nmea\'==options.ioType:\n')
    o.write('\t\t\t    bv = binary.ais6tobitvec(msg.split(\',\')[5])\n')
    o.write('\t\t\t    '+printFields+'('+decodeFunction+'(bv))\n')
    o.write('\t\t\telse: sys.exit(\'ERROR: unknown ioType.  Help!\')\n')
    o.write('\n')
    



######################################################################
if __name__=='__main__':
    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)

    myparser.add_option('-o','--output',dest='outputFileName',default=None,
			help='Name of the python file to write')
#			help='Name of the python file to write [default: stdout]')

    myparser.add_option('-i','-x','--xml-definition',dest='xmlFileName',default=None,
			help='XML definition file for the msg to use')
#			help='XML definition file for the msg to use [default: stdin]')

#    myparser.add_option('-m','--message',dest='message',default=None,
#			help='Which message to write code for [default: all]')

    myparser.add_option('--doc-test',dest='doctest',default=False,action='store_true',
                        help='run the documentation tests')

    myparser.add_option('-p','--prefix',dest='prefix',default=False,action='store_true',
                        help='put the field name in front of all function names.'
			+'  Allows multiple messages in one file')

    myparser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
                        help='run the tests run in verbose mode')

    (options,args) = myparser.parse_args()

    success=True

    if options.doctest:
	import os; print os.path.basename(sys.argv[0]), 'doctests ...',
	argvOrig = sys.argv
	sys.argv= [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	import doctest
	numfail,numtests=doctest.testmod()
	if numfail==0: print 'ok'
	else: 
	    print 'FAILED'
	    success=False
	sys.argv = argvOrig # Restore the original args
	del argvOrig # hide from epydoc
	sys.exit() # FIX: Will this exit success?

    #infile=sys.stdin
    #if options.xmlFileName: infile = file(options.xmlFileName,'r')

    #outfile=sys.stdout
    #if options.outputFile: outfile = file(options.xmlDefinition,'w')

    # FIX: down the road, it would be good to allow either filenames of std{in/out}

    if None==options.xmlFileName:
	sys.exit('ERROR: must specify an xml definition file.')
    if None==options.outputFileName:
	sys.exit('ERROR: must specify an python file to write to.')
    generatePython(options.xmlFileName,options.outputFileName,prefixName=options.prefix)

    print '\nrecommend running pychecker like this:'
    print '  pychecker -q',options.outputFileName
